@@ -1,6 +1,5 @@
-from datetime import datetime, timedelta
+from datetime import datetime
 from dateutil.parser import parse
-import time
 from redis.client import Lock
 
 from limpyd import fields
@@ -8,30 +7,15 @@ from limpyd.contrib import database, collection
 from limpyd_extensions import related
 
 from limpyd_jobs import STATUSES, LimpydJobsException
+from limpyd_jobs.utils import (
+                               datetime_to_score,
+                               compute_delayed_until,
+                               import_class,
+                               )
 
 __all__ = ('BaseJobsModel', 'Queue', 'Job', 'Error')
 
 
-def datetime_to_score(dt):
-    """
-    Convert the given datetime object to be usable as a zset score
-    """
-    return time.mktime(dt.timetuple()) + dt.microsecond / 1000000.0
-
-
-def get_delta(value):
-    """
-    Return a timedelta object based on the value which can be a timedelta
-    or a number of seconds (int or float).
-    Raise an exception in all other cases.
-    """
-    if isinstance(value, (int, float)):
-        return timedelta(seconds=value)
-    elif isinstance(value, timedelta):
-        return value
-    raise Exception('Invalid delta')
-
-
 class BaseJobsModel(related.RelatedModel):
     collection_manager = collection.ExtendedCollectionManager
     database = database.PipelineDatabase()
@@ -72,49 +56,79 @@ class Queue(BaseJobsModel):
 
         return queue
 
-    def delay_job(self, job_pk, delayed_until):
+    def delay_job(self, job, delayed_until):
         """
         Add the job to the delayed list (zset) of the queue.
         """
         timestamp = datetime_to_score(delayed_until)
-        self.delayed.zadd(timestamp, job_pk)
+        self.delayed.zadd(timestamp, job.ident)
 
-    def enqueue_job(self, job_pk, prepend=False):
+    def enqueue_job(self, job, prepend=False):
         """
         Add the job to the waiting list, at the end (it's a fifo list). If
         `prepend` is True, add it at the beginning of the list.
         """
         push_method = getattr(self.waiting, 'lpush' if prepend else 'rpush')
-        push_method(job_pk)
+        push_method(job.ident)
+
+    @staticmethod
+    def _get_iterable_for_names(names):
+        """
+        Ensure that we have an iterable list of names, even if we have a single
+        name
+        """
+        if isinstance(names, basestring):
+            names = (names, )
+        return names
 
     @classmethod
-    def get_all_by_priority(cls, name):
+    def get_all(cls, names):
         """
-        Return all the queues with the given name, sorted by priorities (higher
-        priority first)
+        Return all queues for the given names (for all available priorities)
         """
-        return cls.collection(name=name).sort(by='-priority').instances()
+        names = cls._get_iterable_for_names(names)
+
+        queues = []
+        for queue_name in names:
+            queues.extend(cls.collection(name=queue_name).instances())
+
+        return queues
 
     @classmethod
-    def get_waiting_keys(cls, name):
+    def get_all_by_priority(cls, names):
+        """
+        Return all the queues with the given names, sorted by priorities (higher
+        priority first), then by name
+        """
+        names = cls._get_iterable_for_names(names)
+
+        queues = cls.get_all(names)
+
+        # sort all queues by priority
+        queues.sort(key=lambda q: int(q.priority.hget() or 0), reverse=True)
+
+        return queues
+
+    @classmethod
+    def get_waiting_keys(cls, names):
         """
         Return a list of all queue waiting keys, to use with blpop
         """
-        return [col.waiting.key for col in cls.get_all_by_priority(name)]
+        return [queue.waiting.key for queue in cls.get_all_by_priority(names)]
 
     @classmethod
-    def count_waiting_jobs(cls, name):
+    def count_waiting_jobs(cls, names):
         """
-        Return the number of all jobs waiting in queues with the given name
+        Return the number of all jobs waiting in queues with the given names
         """
-        return sum([col.waiting.llen() for col in cls.get_all_by_priority(name)])
+        return sum([queue.waiting.llen() for queue in cls.get_all(names)])
 
     @classmethod
-    def count_delayed_jobs(cls, name):
+    def count_delayed_jobs(cls, names):
         """
-        Return the number of all delayed jobs in queues with the given name
+        Return the number of all delayed jobs in queues with the given names
         """
-        return sum([col.delayed.zcard() for col in cls.get_all_by_priority(name)])
+        return sum([queue.delayed.zcard() for queue in cls.get_all(names)])
 
     @property
     def first_delayed(self):
@@ -136,7 +150,7 @@ class Queue(BaseJobsModel):
 
         return first_entry[1] if first_entry else None
 
-    def requeue_delayed_jobs(self, job_model):
+    def requeue_delayed_jobs(self):
         """
         Put all delayed jobs that are now ready, back in the queue waiting list
         """
@@ -175,7 +189,7 @@ class Queue(BaseJobsModel):
                     break
 
                 # split into vars for readability
-                job_pk, delayed_until = first_entry
+                job_ident, delayed_until = first_entry
 
                 # if the date of the job is in the future, another work took the
                 # job we wanted, so we let this job here and stop the loop as we
@@ -184,12 +198,12 @@ class Queue(BaseJobsModel):
                     break
 
                 # remove the entry we just got from the delayed ones
-                self.delayed.zrem(job_pk)
+                self.delayed.zrem(job_ident)
 
                 # and add it to the waiting queue
-                job = job_model.get(job_pk)
+                job = Job.get_from_ident(job_ident)
                 job.status.hset(STATUSES.WAITING)
-                self.enqueue_job(job_pk)
+                self.enqueue_job(job)
 
 
 class Job(BaseJobsModel):
@@ -203,9 +217,46 @@ class Job(BaseJobsModel):
     delayed_until = fields.InstanceHashField()
 
     queue_model = Queue
+    queue_name = None
 
     @classmethod
-    def add_job(cls, identifier, queue_name, priority=0, queue_model=None,
+    def get_model_repr(cls):
+        """
+        Return a string representation of the current class
+        """
+        return '%s.%s' % (cls.__module__, cls.__name__)
+
+    @property
+    def ident(self):
+        """
+        Return the string to use in  queues, include the job's class and its pk
+        """
+        return '%s:%s' % (self.get_model_repr(), self.pk.get())
+
+    @classmethod
+    def get_from_ident(self, ident):
+        """
+        Take a string as returned by get_ident and return a job,
+        based on the class representation and the job's pk from the ident
+        """
+        model_repr, job_pk = ident.split(':', 1)
+        klass = import_class(model_repr)
+        return klass.get(job_pk)
+
+    @classmethod
+    def _get_queue_name(cls, queue_name=None):
+        """
+        Return the given queue_name if defined, else the class's one.
+        If both are None, raise an Exception
+        """
+        if queue_name is None and cls.queue_name is None:
+            raise LimpydJobsException("Queue's name not defined")
+        if queue_name is None:
+            return cls.queue_name
+        return queue_name
+
+    @classmethod
+    def add_job(cls, identifier, queue_name=None, priority=0, queue_model=None,
                 prepend=False, delayed_for=None, delayed_until=None,
                 **fields_if_new):
         """
@@ -221,21 +272,14 @@ class Job(BaseJobsModel):
         """
 
         # check for delayed_for/delayed_until arguments
-        if delayed_until:
-            if delayed_for:
-                raise ValueError('delayed_until and delayed_for arguments are exclusive')
-            if not isinstance(delayed_until, datetime):
-                raise ValueError('Invalid delayed_until argument: must be a datetime object')
-
-        if delayed_for:
-            try:
-                delayed_until = datetime.utcnow() + get_delta(delayed_for)
-            except Exception:
-                raise ValueError('Invalid delayed_for argument: must be an int, a float or a timedelta object')
+        delayed_until = compute_delayed_until(delayed_for, delayed_until)
 
         # create the job or get an existing one
         job, created = cls.get_or_connect(identifier=identifier, status=STATUSES.WAITING)
 
+        # check queue_name
+        queue_name = cls._get_queue_name(queue_name)
+
         # if the job already exists, and we want a higher priority or move it,
         # start by updating it
         if not created:
@@ -252,7 +296,7 @@ class Job(BaseJobsModel):
             if queue_model is None:
                 queue_model = cls.queue_model
             current_queue = queue_model.get_queue(queue_name, current_priority)
-            current_queue.waiting.lrem(0, job.pk.get())
+            current_queue.waiting.lrem(0, job.ident)
 
         elif fields_if_new:
             job.set_fields(added=str(datetime.utcnow()), **fields_if_new)
@@ -262,6 +306,15 @@ class Job(BaseJobsModel):
 
         return job
 
+    def run(self, queue):
+        """
+        The method called by the work to do some stuff. Must be overriden to
+        your needs.
+        The optional return value of this function will be passed to the
+        job_success method of the worker.
+        """
+        raise NotImplementedError('You must implement your own action')
+
     @property
     def duration(self):
         """
@@ -274,10 +327,13 @@ class Job(BaseJobsModel):
         except:
             return None
 
-    def requeue(self, queue_name, priority=None, requeue_delay_delta=None, queue_model=None):
+    def requeue(self, queue_name=None, priority=None, delayed_for=None,
+                                        delayed_until=None, queue_model=None):
         """
         Requeue the job in the given queue if it has previously failed
         """
+        queue_name = self._get_queue_name(queue_name)
+
         # we can only requeue a job that raised an error
         if self.status.hget() != STATUSES.ERROR:
             raise LimpydJobsException('Job cannot be requeued if not in ERROR status')
@@ -287,17 +343,24 @@ class Job(BaseJobsModel):
         if priority is None:
             priority = self.priority.hget()
 
-        delayed_until = None
-        if requeue_delay_delta:
-            delayed_until = datetime.utcnow() + get_delta(requeue_delay_delta)
+        delayed_until = compute_delayed_until(delayed_for, delayed_until)
 
         self.enqueue_or_delay(queue_name, priority, delayed_until, queue_model=queue_model)
 
-    def enqueue_or_delay(self, queue_name, priority, delayed_until=None, prepend=False, queue_model=None):
+    def enqueue_or_delay(self, queue_name=None, priority=None,
+                         delayed_until=None, prepend=False, queue_model=None):
         """
         Will enqueue or delay the job depending of the delayed_until.
         """
-        fields = {'priority': priority}
+        queue_name = self._get_queue_name(queue_name)
+
+        fields = {}
+
+        if priority is not None:
+            fields['priority'] = priority
+        else:
+            priority = self.priority.hget()
+
         in_the_future = delayed_until and delayed_until > datetime.utcnow()
 
         if in_the_future:
@@ -314,12 +377,31 @@ class Job(BaseJobsModel):
         queue = queue_model.get_queue(queue_name, priority)
 
         if in_the_future:
-            queue.delay_job(self.pk.get(), delayed_until)
+            queue.delay_job(self, delayed_until)
         else:
-            queue.enqueue_job(self.pk.get(), prepend)
+            queue.enqueue_job(self, prepend)
+
+    # Methods below will be called only if defined, so we don't create them, but
+    # feel free to do so if you need to interact on jobs updates
+
+    # def on_started(self, queue):
+    #     pass
+
+    # def on_success(self, queue, result):
+    #     pass
+
+    # def on_error(self, queue, exception, trace):
+    #     pass
+
+    # def on_skipped(self, queue):
+    #     pass
+
+    # def on_requeued(self, queue):
+    #     pass
 
 
 class Error(BaseJobsModel):
+    job_model_repr = fields.InstanceHashField(indexable=True)
     job_pk = fields.InstanceHashField(indexable=True)
     identifier = fields.InstanceHashField(indexable=True)
     queue_name = fields.InstanceHashField(indexable=True)
@@ -348,6 +430,7 @@ class Error(BaseJobsModel):
 
         fields = dict(
             queue_name=queue_name,
+            job_model_repr=job.get_model_repr(),
             job_pk=job.pk.get(),
             identifier=getattr(job, '_cached_identifier', job.identifier.hget()),
             date=str(when.date()),
