@@ -14,28 +14,30 @@ from limpyd.exceptions import DoesNotExist
 
 from limpyd_jobs import STATUSES, LimpydJobsException, ConfigurationException
 from limpyd_jobs.models import Queue, Job, Error
+from limpyd_jobs.utils import import_class, total_seconds
 
-LOGGER_BASE_NAME = 'limpyd-jobs'
-logger = logging.getLogger(LOGGER_BASE_NAME)
+LOGGER_NAME = 'limpyd-jobs'
+logger = logging.getLogger(LOGGER_NAME)
 
 
 class Worker(object):
-    # name of the worker (must be name of Queue objets)
-    name = None
+    # names of queues to use
+    queues = None
 
-    # models to use
+    # models to use (for all queues/errors !)
     queue_model = Queue
-    job_model = Job
     error_model = Error
 
     # logging information
-    logger_base_name = LOGGER_BASE_NAME + '.%s'  # will use self.name
+    logger_name = LOGGER_NAME
     logger_level = logging.ERROR
     save_errors = True
     save_tracebacks = True
 
     # maximum number of loops to run
     max_loops = 1000
+    # max total duration of the worker
+    max_duration = None
     # max delay for blpop
     timeout = 30
     # minimum time in seconds between the update of keys to fetch
@@ -54,62 +56,64 @@ class Worker(object):
     # we want to intercept SIGTERM and SIGINT signals, to stop gracefuly
     terminate_gracefuly = True
 
-    def __init__(self, name=None, callback=None,
-                 queue_model=None, job_model=None, error_model=None,
-                 logger_base_name=None, logger_level=None, save_errors=None,
-                 save_tracebacks=None, max_loops=None, terminate_gracefuly=None,
-                 timeout=None, fetch_priorities_delay=None,
-                 fetch_delayed_delay=None, requeue_times=None,
-                 requeue_priority_delta=None, requeue_delay_delta=None):
+    # callback to run for each job, will be set to self.execute if None
+    callback = None
+
+    # all parameters that can be passed to the constructor
+    parameters = ('queues', 'callback', 'queue_model', 'error_model',
+                  'logger_name', 'logger_level', 'save_errors',
+                  'save_tracebacks', 'max_loops', 'max_duration',
+                  'terminate_gracefuly', 'timeout', 'fetch_priorities_delay',
+                  'fetch_delayed_delay', 'requeue_times',
+                  'requeue_priority_delta', 'requeue_delay_delta')
+
+    @staticmethod
+    def _parse_queues(queues):
+        """
+        Parse the given parameter and return a list of queues.
+        The parameter must be a list/tuple of strings, or a string with queues
+        separated by a comma.
+        """
+        if not queues:
+            raise ConfigurationException('The queue(s) to use are not defined')
+
+        if isinstance(queues, (list, tuple)):
+            for queue_name in queues:
+                if not isinstance(queue_name, basestring):
+                    raise ConfigurationException('Queue name "%s" is not a (base)string')
+
+        elif isinstance(queues, basestring):
+            queues = queues.split(',')
+
+        else:
+            raise ConfigurationException('Invalid format for queues names')
+
+        return list(queues)
+
+    def __init__(self, queues=None, **kwargs):
         """
         Create the worker by saving arguments, doing some checks, preparing
         logger and signals management, and getting queues keys.
         """
-        if name is not None:
-            self.name = name
+        if queues is not None:
+            self.queues = queues
+        self.queues = self._parse_queues(self.queues)
 
-        if not self.name:
-            raise ConfigurationException('The name of the worker is not defined')
+        for parameter in self.parameters:
+            if parameter in kwargs:
+                setattr(self, parameter, kwargs[parameter])
 
         # save and check models to use
-        if queue_model is not None:
-            self.queue_model = queue_model
         self._assert_correct_model(self.queue_model, Queue, 'queue')
-        if job_model is not None:
-            self.job_model = job_model
-        self._assert_correct_model(self.job_model, Job, 'job')
-        if error_model is not None:
-            self.error_model = error_model
         self._assert_correct_model(self.error_model, Error, 'error')
 
-        # process other arguments
-        self.callback = callback if callback is not None else self.execute
-        if max_loops is not None:
-            self.max_loops = max_loops
-        if terminate_gracefuly is not None:
-            self.terminate_gracefuly = terminate_gracefuly
-        if save_errors is not None:
-            self.save_errors = save_errors
-        if save_tracebacks is not None:
-            self.save_tracebacks = save_tracebacks
-        if timeout is not None:
-            self.timeout = timeout
-        if fetch_priorities_delay is not None:
-            self.fetch_priorities_delay = fetch_priorities_delay
-        if fetch_delayed_delay is not None:
-            self.fetch_delayed_delay = fetch_delayed_delay
-        if requeue_times is not None:
-            self.requeue_times = requeue_times
-        if requeue_priority_delta is not None:
-            self.requeue_priority_delta = requeue_priority_delta
-        if requeue_delay_delta is not None:
-            self.requeue_delay_delta = requeue_delay_delta
+        # process other special arguments
+        if not self.callback:
+            self.callback = self.execute
+        if self.max_duration:
+            self.max_duration = timedelta(seconds=self.max_duration)
 
         # prepare logging
-        if logger_base_name is not None:
-            self.logger_base_name = logger_base_name
-        if logger_level is not None:
-            self.logger_level = logger_level
         self.set_logger()
 
         if self.terminate_gracefuly:
@@ -117,6 +121,9 @@ class Worker(object):
 
         self.keys = []  # list of redis keys to listen
         self.num_loops = 0  # loops counter
+        self.start_date = None  # set when run is called
+        self.end_date = None  # set when run ends
+        self.wanted_end_date = None  # will be the end after which the worker must stop
         self.end_forced = False  # set it to True in "execute" to force stop just after
         self.status = None  # is set to None/waiting/running by the worker
         self.end_signal_caught = False  # internaly set to True if end signal caught
@@ -158,9 +165,9 @@ class Worker(object):
 
     def set_logger(self):
         """
-        Prepare the logger, based on self.logger_base_name and self.logger_level
+        Prepare the logger, using self.logger_name and self.logger_level
         """
-        self.logger = logging.getLogger(self.logger_base_name % self.name)
+        self.logger = logging.getLogger(self.logger_name)
         self.logger.setLevel(self.logger_level)
 
     @property
@@ -189,8 +196,9 @@ class Worker(object):
         """
         Return True if the worker must stop when the current loop is over.
         """
-        return self.terminate_gracefuly and self.end_signal_caught \
-            or self.num_loops >= self.max_loops or self.end_forced
+        return bool(self.terminate_gracefuly and self.end_signal_caught
+                 or self.num_loops >= self.max_loops or self.end_forced
+                 or self.wanted_end_date and datetime.utcnow() >= self.wanted_end_date)
 
     def set_status(self, status):
         """
@@ -213,15 +221,15 @@ class Worker(object):
         blpop_result = self.connection.blpop(self.keys, self.timeout)
         if blpop_result is None:
             return None
-        queue_redis_key, job_pk = blpop_result
+        queue_redis_key, job_ident = blpop_result
         self.set_status('running')
-        return self.get_queue(queue_redis_key), self.get_job(job_pk)
+        return self.get_queue(queue_redis_key), self.get_job(job_ident)
 
-    def get_job(self, job_pk):
+    def get_job(self, job_ident):
         """
-        Return a job based on its primary key.
+        Return a job based on its representation (class + pk).
         """
-        return self.job_model.get(job_pk)
+        return Job.get_from_ident(job_ident)
 
     def get_queue(self, queue_redis_key):
         """
@@ -262,26 +270,29 @@ class Worker(object):
         The optional return value of this function will be passed to the
         job_success method.
         """
-        raise NotImplementedError('You must implement your own action')
+        return job.run(queue)
 
     def update_keys(self):
         """
         Update the redis keys to listen for new jobs priorities.
         """
-        self.keys = self.queue_model.get_waiting_keys(self.name)
+        self.keys = self.queue_model.get_waiting_keys(self.queues)
+
+        if not self.keys:
+            self.log('No queues yet', level='warning')
         self.last_update_keys = datetime.utcnow()
 
     def count_waiting_jobs(self):
         """
         Return the number of all jobs waiting in queues managed by the worker
         """
-        return self.queue_model.count_waiting_jobs(self.name)
+        return self.queue_model.count_waiting_jobs(self.queues)
 
     def count_delayed_jobs(self):
         """
         Return the number of all delayed jobs in queues managed by the worker
         """
-        return self.queue_model.count_delayed_jobs(self.name)
+        return self.queue_model.count_delayed_jobs(self.queues)
 
     def run_started(self):
         """
@@ -302,36 +313,50 @@ class Worker(object):
 
         self.set_status('starting')
 
-        must_stop = False
+        self.start_date = datetime.utcnow()
+        if self.max_duration:
+            self.wanted_end_date = self.start_date + self.max_duration
 
-        # get keys or wait for queues available if no ones are yet
-        self.update_keys()
-        self.requeue_delayed_jobs()
-
-        if not self.keys:
-            self.log('No queues yet with the name %s.' % self.name, level='warning')
-            while not self.keys:
-                sleep(self.fetch_priorities_delay)
-                must_stop = self.must_stop()
-                if must_stop:
-                    break
-                self.update_keys()
+        must_stop = self.must_stop()
 
         if not must_stop:
-            self.run_started()
-            self._main_loop()
+
+            # get keys
+            while not self.keys and not must_stop:
+                self.update_keys()
+                if not self.keys:
+                    sleep(self.fetch_priorities_delay)
+                must_stop = self.must_stop()
+
+            if not must_stop:
+                # wait for queues available if no ones are yet
+                self.requeue_delayed_jobs()
+                self.run_started()
+                self._main_loop()
 
         self.set_status('terminated')
+        self.end_date = datetime.utcnow()
         self.run_ended()
         if self.terminate_gracefuly:
             self.stop_handling_end_signal()
 
+    @property
+    def elapsed(self):
+        """
+        Return a timedelta representation of the time passed sine the worker
+        was running.
+        """
+        if not self.start_date:
+            return None
+        return (self.end_date or datetime.utcnow()) - self.start_date
+
     def requeue_delayed_jobs(self):
         """
         Requeue each delayed job that are now ready to be executed
         """
-        for queue in self.queue_model.get_all_by_priority(self.name):
-            queue.requeue_delayed_jobs(job_model=self.job_model)
+        for queue in self.queue_model.get_all_by_priority(self.queues):
+            queue.requeue_delayed_jobs()
+
         self.last_requeue_delayed = datetime.utcnow()
 
     def _main_loop(self):
@@ -360,7 +385,10 @@ class Worker(object):
                 self.log('Unable to get job: %s' % str(e), level='error')
             else:
                 self.num_loops += 1
-                identifier = 'pk:%s' % job.pk.get()  # default if failure
+                try:
+                    identifier = 'pk:%s' % job.pk.get()
+                except Exception, e:
+                    identifier = '??'
                 try:
                     self.set_status('running')
                     identifier, status = job.hmget('identifier', 'status')
@@ -384,13 +412,19 @@ class Worker(object):
                             self.job_success(job, queue, job_result)
                 except Exception, e:
                     self.log('[%s] unexpected error: %s' % (identifier, str(e)),
-                             level='error')
+                                                                 level='error')
+                    try:
+                        queue.errors.rpush(job.ident)
+                    except Exception, e:
+                        self.log('[%s] unable to add the error in the queue: %s'
+                         % (identifier, str(e)), level='error')
 
     def run_ended(self):
         """
         Called just after ending the run loop. Actually only do logging.
         """
-        self.log('Run terminated, with %d loops.' % self.num_loops)
+        self.log('Run terminated, with %d loops (duration=%s)' % (
+                                                self.num_loops, self.elapsed))
 
     def additional_error_fields(self, job, queue, exception, trace=None):
         """
@@ -403,7 +437,7 @@ class Worker(object):
         Called when an exception was raised during the execute call for a job.
         """
         job.hmset(end=str(datetime.utcnow()), status=STATUSES.ERROR)
-        queue.errors.rpush(job.pk.get())
+        queue.errors.rpush(job.ident)
 
         if self.save_errors:
             additional_fields = self.additional_error_fields(job, queue, exception)
@@ -415,12 +449,31 @@ class Worker(object):
 
         self.log(self.job_error_message(job, queue, exception, trace), level='error')
 
+        if hasattr(job, 'on_error'):
+            job.on_error(queue, exception, trace)
+
+        # requeue the job if needed
         if self.requeue_times and self.requeue_times >= int(job.tries.hget() or 0):
-            name, priority = queue.hmget('name', 'priority')
+            priority = queue.priority.hget()
+
             if self.requeue_priority_delta:
                 priority = int(priority) + self.requeue_priority_delta
-            job.requeue(name, priority, self.requeue_delay_delta, self.queue_model)
-            self.log(self.job_requeue_message(job, queue, priority))
+
+            self.requeue_job(job, queue, priority, delayed_for=self.requeue_delay_delta)
+
+    def requeue_job(self, job, queue, priority, delayed_for=None):
+        """
+        Requeue a job in a queue with the given priority, possibly delayed
+        """
+        job.requeue(queue_name=queue._cached_name,
+                    priority=priority,
+                    delayed_for=delayed_for,
+                    queue_model=self.queue_model)
+
+        if hasattr(job, 'on_requeued'):
+            job.on_requeued(queue)
+
+        self.log(self.job_requeue_message(job, queue))
 
     def job_error_message(self, job, queue, exception, trace=None):
         """
@@ -429,15 +482,15 @@ class Worker(object):
         return '[%s|%s] error: %s' % (queue._cached_name,
                                       job._cached_identifier, str(exception))
 
-    def job_requeue_message(self, job, queue, priority):
+    def job_requeue_message(self, job, queue):
         """
         Return the message to log when a job is requeued
         """
+        priority, delayed_until = job.hmget('priority', 'delayed_until')
 
-        msg = '[%s|%s] requeued with priority %'
-        args = [queue._cached_name, job._cached_identifier]
+        msg = '[%s|%s] requeued with priority %s'
+        args = [queue._cached_name, job._cached_identifier, priority]
 
-        delayed_until = job.delayed_until.hget()
         if delayed_until:
             msg += ', delayed until %s'
             args.append(delayed_until)
@@ -450,8 +503,10 @@ class Worker(object):
         job_result is the value returned by the callback, if any.
         """
         job.hmset(end=str(datetime.utcnow()), status=STATUSES.SUCCESS)
-        queue.success.rpush(job.pk.get())
+        queue.success.rpush(job.ident)
         self.log(self.job_success_message(job, queue, job_result))
+        if hasattr(job, 'on_success'):
+            job.on_success(queue, job_result)
 
     def job_success_message(self, job, queue, job_result):
         """
@@ -467,6 +522,8 @@ class Worker(object):
         job.hmset(start=str(datetime.utcnow()), status=STATUSES.RUNNING)
         job.tries.hincrby(1)
         self.log(self.job_started_message(job, queue))
+        if hasattr(job, 'on_started'):
+            job.on_started(queue)
 
     def job_started_message(self, job, queue):
         """
@@ -479,6 +536,8 @@ class Worker(object):
         Called if a job can't be run: canceled, already running or done.
         """
         self.log(self.job_skipped_message(job, queue), level='warning')
+        if hasattr(job, 'on_skipped'):
+            job.on_skipped(queue)
 
     def job_skipped_message(self, job, queue):
         """
@@ -512,19 +571,17 @@ class WorkerConfig(object):
                   'default to limpyd_jobs.workers.WorkerConfig'),
 
         make_option('--print-options', action='store_true', dest='print_options',
-            help='Print options as parsed by the script, e.g. --print-options'),
+            help='Print options used by the worker, e.g. --print-options'),
         make_option('--dry-run', action='store_true', dest='dry_run',
             help='Won\'t execute any job, just starts the worker and finish it immediatly, e.g. --dry-run'),
 
-        make_option('--name', action='store', dest='name',
-            help='Name of the Queues to handle e.g. --name=my-queue-name'),
+        make_option('--queues', action='store', dest='queues',
+            help='Name of the Queues to handle, comma separated e.g. --queues=queue1,queue2'),
 
-        make_option('--job-model', action='store', dest='job_model',
-            help='Name of the Job model to use, e.g. --job-model=my.module.JobModel'),
         make_option('--queue-model', action='store', dest='queue_model',
-            help='Name of the Queue model to use, e.g. --queue-model=my.module.QueueModel'),
+            help='Name of the Queue model to use (for all queues !), e.g. --queue-model=my.module.QueueModel'),
         make_option('--error-model', action='store', dest='error_model',
-            help='Name of the Error model to use, e.g. --queue-model=my.module.ErrorModel'),
+            help='Name of the Error model to use (for all errors !), e.g. --queue-model=my.module.ErrorModel'),
 
         make_option('--worker-class', action='store', dest='worker_class',
             help='Name of the Worker class to use, e.g. --worker-class=my.module.WorkerClass'),
@@ -532,8 +589,8 @@ class WorkerConfig(object):
         make_option('--callback', action='store', dest='callback',
             help='The callback to call for each job, e.g. --worker-class=my.module.callback'),
 
-        make_option('--logger-base-name', action='store', dest='logger_base_name',
-            help='The base name to use for logging, e.g. --logger-base-name="limpyd-jobs.%s"'),
+        make_option('--logger-name', action='store', dest='logger_name',
+            help='The base name to use for logging, e.g. --logger-name="limpyd-jobs.my-job"'),
         make_option('--logger-level', action='store', dest='logger_level',
             help='The level to use for logging, e.g. --worker-class=INFO'),
 
@@ -548,6 +605,8 @@ class WorkerConfig(object):
 
         make_option('--max-loops', type='int', dest='max_loops',
             help='Max number of jobs to run, e.g. --max-loops=100'),
+        make_option('--max-duration', type='int', dest='max_duration',
+            help='Max duration of the worker, in seconds (None by default), e.g. --max-duration=3600'),
 
         make_option('--terminate-gracefuly', action='store_true', dest='terminate_gracefuly',
             help='Intercept SIGTERM and SIGINT signals to stop gracefuly, e.g. --terminate-gracefuly'),
@@ -580,52 +639,12 @@ class WorkerConfig(object):
     )
 
     default_classes = {
-        'job_model': Job,
         'queue_model': Queue,
         'error_model': Error,
         'worker_class': Worker,
         'callback': None,
     }
 
-    worker_options = ('name', 'job_model', 'queue_model', 'error_model',
-                      'callback', 'logger_base_name', 'logger_level',
-                      'save_errors', 'save_tracebacks', 'max_loops',
-                      'terminate_gracefuly', 'timeout',
-                      'fetch_priorities_delay', 'fetch_delayed_delay',
-                      'requeue_times', 'requeue_priority_delta',
-                      'requeue_delay_delta')
-
-    @staticmethod
-    def _import_module(module_uri):
-        """
-        Replacement to import_module from importlib for python 2.6
-        """
-        return __import__(module_uri, {}, {}, [''])  # pragma: no cover
-
-    @staticmethod
-    def import_class(class_uri):
-        """
-        Import a class by string 'from.path.module.class'
-        To support python 2.6
-        """
-        try:
-            from importlib import import_module  # pragma: no cover
-            callback = import_module  # pragma: no cover
-        except ImportError:  # pragma: no cover
-            callback = WorkerConfig._import_module  # pragma: no cover
-
-        parts = class_uri.split('.')
-        class_name = parts.pop()
-        module_uri = '.'.join(parts)
-
-        try:
-            module = callback(module_uri)
-        except ImportError:
-            # maybe we are still in a module, test going up one level
-            module = WorkerConfig.import_class(module_uri)
-
-        return getattr(module, class_name)
-
     def __init__(self, argv=None):
         """
         Save arguments and program name
@@ -633,8 +652,6 @@ class WorkerConfig(object):
         self.argv = argv or sys.argv[:]
         self.prog_name = os.path.basename(self.argv[0])
         self.manage_options()
-        if self.options.print_options:
-            self.print_options()
         self.update_proc_title()
 
     def get_version(self):
@@ -687,7 +704,10 @@ class WorkerConfig(object):
                     self.parser.error('Invalid logger-level %s' % self.options.logger_level)
 
         if self.options.max_loops is not None and self.options.max_loops < 0:
-            self.parser.error('The max-loops argument (%s) must be a positive integer' % self.options.max_loops)
+            self.parser.error('The max-loops argument (%s) must be a <positive></positive> integer' % self.options.max_loops)
+
+        if self.options.max_duration is not None and self.options.max_duration < 0:
+            self.parser.error('The max-duration argument (%s) must be a positive integer' % self.options.max_duration)
 
         if self.options.timeout is not None and self.options.timeout < 0:
             self.parser.error('The timeout argument (%s) must be a positive integer (including 0)' % self.options.timeout)
@@ -717,7 +737,7 @@ class WorkerConfig(object):
         """
         option = getattr(self.options, name)
         if option:
-            klass = WorkerConfig.import_class(option)
+            klass = import_class(option)
         else:
             klass = self.default_classes[name]
         setattr(self.options, name, klass)
@@ -738,30 +758,30 @@ class WorkerConfig(object):
         """
         options = []
 
-        if self.__class__ != WorkerConfig:
-            options.append(("worker_config", '%s.%s' % (self.__class__.__module__,
-                                                        self.__class__.__name__)))
+        print "The script is running with the following options:"
 
-        if self.database_config:
-            options.append(("database", '%s:%s:%s' % (self.database_config['host'],
-                                                      self.database_config['port'],
-                                                      self.database_config['db'])))
+        options.append(("dry_run", self.options.dry_run))
 
-        if self.options.dry_run:
-            options.append(("dry_run", self.options.dry_run))
+        options.append(("worker_config", self.__class__))
 
-        for option_name in self.worker_options:
-            option = getattr(self.options, option_name)
-            if option is not None:
-                options.append((option_name.replace('_', '-'), option))
-            # special case for worker_class which is not a worker option
-            if option_name == 'error_model' and self.options.worker_class is not None:
-                options.append(("worker-class", self.options.worker_class))
+        database_config = self.database_config or \
+                          self.options.queue_model.database.connection_settings
+        options.append(("database", '%s:%s:%s' % (database_config['host'],
+                                                  database_config['port'],
+                                                  database_config['db'])))
 
-        if options:
-            print "The worker will run with the following options:"
-            for name, value in options:
-                print " - %s = %s" % (name, value)
+        if self.options.worker_class is not None:
+            options.append(("worker-class", self.options.worker_class))
+
+        for name, value in options:
+            print " - %s = %s" % (name.replace('_', '-'), value)
+
+        print "The worker will run with the following options:"
+        for name in self.options.worker_class.parameters:
+            option = getattr(self.worker, name)
+            if isinstance(option, (list, tuple, set)):
+                option = ','.join(option)
+            print " - %s = %s" % (name.replace('_', '-'), option)
 
     def execute(self):
         """
@@ -769,6 +789,8 @@ class WorkerConfig(object):
         """
         self.prepare_models()
         self.prepare_worker()
+        if self.options.print_options:
+            self.print_options()
         self.run()
 
     def prepare_models(self):
@@ -776,8 +798,7 @@ class WorkerConfig(object):
         If a database config ws given as argument, apply it to our models
         """
         if self.database_config:
-            for model in (self.options.job_model, self.options.queue_model,
-                                                    self.options.error_model):
+            for model in (self.options.queue_model, self.options.error_model):
                 model.database.reset(**self.database_config)
 
     def prepare_worker_options(self):
@@ -785,7 +806,7 @@ class WorkerConfig(object):
         Prepare (and return as a dict) all options to be passed to the worker
         """
         worker_options = dict()
-        for option_name in (self.worker_options):
+        for option_name in (self.options.worker_class.parameters):
             option = getattr(self.options, option_name)
             if option is not None:
                 worker_options[option_name] = option
@@ -831,8 +852,8 @@ class WorkerConfig(object):
                 status += ' - ending'
         title_parts.append('[%s]' % status)
 
-        if has_worker and self.worker.name:
-            title_parts.append('queue=%s' % self.worker.name)
+        if has_worker and self.worker.queues:
+            title_parts.append('queues=%s' % ','.join(self.worker.queues))
 
         if has_worker and self.worker.status:
             # add infos about the main loop
@@ -844,6 +865,15 @@ class WorkerConfig(object):
             # and about the number of delayed jobs
             title_parts.append('delayed=%s' % self.worker.count_delayed_jobs())
 
+            # and about elapsed time
+            if self.worker.start_date:
+                duraiton_message = 'duration=%s'
+                duration_args = (timedelta(seconds=int(round(total_seconds(self.worker.elapsed)))), )
+                if self.worker.max_duration:
+                    duraiton_message += '/%s'
+                    duration_args += (self.worker.max_duration, )
+                title_parts.append(duraiton_message % duration_args)
+
         return ' '.join(title_parts)
 
     def update_proc_title(self, worker=None):
